@@ -1,10 +1,10 @@
 #******************************************************************************
-# $Id: POP3Client.pm,v 2.1 1999/08/19 19:16:37 sdowd Exp $
+# $Id: POP3Client.pm,v 2.11 2002/06/18 01:57:54 ssd Exp $
 #
 # Description:  POP3Client module - acts as interface to POP3 server
-# Author:       Sean Dowd <dowd@home.com> or <sdowd@arcmail.com>
+# Author:       Sean Dowd <pop3client@dowds.net>
 #
-# Copyright (c) 1999  Sean Dowd.  All rights reserved.
+# Copyright (c) 1999,2000,2001  Sean Dowd.  All rights reserved.
 # This module is free software; you can redistribute it and/or modify
 # it under the same terms as Perl itself.
 #
@@ -13,23 +13,23 @@
 package Mail::POP3Client;
 
 use strict;
-use Carp;
-use IO::Socket;
+use Carp ();
+use IO::Socket qw(SOCK_STREAM);
 
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
 
 require Exporter;
 
-@ISA = qw(Exporter AutoLoader);
+@ISA = qw(Exporter);
 # Items to export into callers namespace by default. Note: do not export
 # names by default without a very good reason. Use EXPORT_OK instead.
 # Do not simply export all your public functions/methods/constants.
 @EXPORT = qw(
-	
+
 );
 
-my $ID =q( $Id: POP3Client.pm,v 2.1 1999/08/19 19:16:37 sdowd Exp $ );
-$VERSION = substr q$Revision: 2.1 $, 10;
+my $ID =q( $Id: POP3Client.pm,v 2.11 2002/06/18 01:57:54 ssd Exp $ );
+$VERSION = substr q$Revision: 2.11 $, 10;
 
 
 # Preloaded methods go here.
@@ -39,11 +39,12 @@ $VERSION = substr q$Revision: 2.1 $, 10;
 #* new Mail::POP3Client( USER => user,
 #*                       PASSWORD => pass,
 #*                       HOST => host,
-#*                       AUTH_MODE => [APOP|PASS],
+#*                       AUTH_MODE => [BEST|APOP|CRAM-MD5|PASS],
 #*                       TIMEOUT => 30,
+#*                       LOCALADDR => 'xxx.xxx.xxx.xxx[:xx]',
 #*                       DEBUG => 1 );
 #* OR (deprecated)
-#* new Mail::POP3Client( user, pass, host [, port, debug, auth_mode])
+#* new Mail::POP3Client( user, pass, host [, port, debug, auth_mode, local_addr])
 #******************************************************************************
 sub new
 {
@@ -59,14 +60,18 @@ sub new
 	      MESG => 'OK',
 	      BANNER => '',
 	      MESG_ID => '',
-	      AUTH_MODE => 'APOP',
+	      AUTH_MODE => 'BEST',
 	      EOL => "\015\012",
 	      TIMEOUT => 60,
+	      STRIPCR => 0,
+	      LOCALADDR => undef,
 	     };
+  $self->{tranlog} = ();
+  $^O =~ /MacOS/i && ($self->{STRIPCR} = 1);
   bless( $self, $classname );
   $self->_init( @_ );
 
-  if ( $self->User() && $self->Pass() )
+  if ( defined($self->User()) && defined($self->Pass()) )
     {
       $self->Connect();
     }
@@ -118,10 +123,12 @@ sub _initOldStyle {
   $debug && $self->Debug( $debug );
   my $auth_mode = shift;
   $auth_mode && ($self->{AUTH_MODE} = $auth_mode);
+  my $localaddr = shift;
+  $localaddr && ($self->{LOCALADDR} = $localaddr);
 }
 
 #******************************************************************************
-#* Is the socket alive?
+#* What version are we?
 #******************************************************************************
 sub Version {
   return $VERSION;
@@ -168,7 +175,7 @@ sub Debug
   my $me = shift;
   my $debug = shift or return $me->{DEBUG};
   $me->{DEBUG} = $debug;
-  
+
 } # end Debug
 
 
@@ -179,7 +186,7 @@ sub Port
 {
   my $me = shift;
   my $port = shift or return $me->{PORT};
-  
+
   $me->{PORT} = $port;
 
 } # end port
@@ -198,6 +205,17 @@ sub Host
   $me->{HOST} = $host;
 } # end host
 
+#******************************************************************************
+#* set the local address
+#******************************************************************************
+sub LocalAddr
+{
+  my $me = shift;
+  my $addr = shift or return $me->{LOCALADDR};
+
+  $me->{LOCALADDR} = $addr;
+}
+
 
 #******************************************************************************
 #* query the socket to use as a file handle
@@ -216,7 +234,7 @@ sub User
   my $me = shift;
   my $user = shift or return $me->{USER};
   $me->{USER} = $user;
-  
+
 } # end User
 
 
@@ -228,12 +246,12 @@ sub Pass
   my $me = shift;
   my $pass = shift or return $me->{PASSWORD};
   $me->{PASSWORD} = $pass;
-  
+
 } # end Pass
 
 
 #******************************************************************************
-#* 
+#*
 #******************************************************************************
 sub Count
 {
@@ -244,7 +262,7 @@ sub Count
   } else {
     return $me->{COUNT};
   }
-    
+
 } # end Count
 
 
@@ -260,12 +278,12 @@ sub Size
   } else {
     return $me->{SIZE};
   }
-  
+
 } # end Size
 
 
 #******************************************************************************
-#* 
+#*
 #******************************************************************************
 sub EOL {
   my $me = shift;
@@ -274,23 +292,35 @@ sub EOL {
 
 
 #******************************************************************************
-#* 
+#*
 #******************************************************************************
 sub Close
 {
   my $me = shift;
-  if ($me->Alive()) {
-    my $s = $me->Socket();
+
+  # only send the QUIT message is the socket is still connected.  Some
+  # POP3 servers close the socket after a failed authentication.  It
+  # is unclear whether the RFC allows this or not, so we'll attempt to
+  # check the condition of the socket before sending data here.
+  if ($me->Alive() && $me->Socket() && $me->Socket()->connected() ) {
     $me->_sockprint( "QUIT", $me->EOL );
+
+    # from Patrick Bourdon - need this because some servers do not
+    # delete in all cases.  RFC says server can respond (in UPDATE
+    # state only, otherwise always OK).
+    my $line = $me->_sockread();
+
+    $me->Message( $line );
     close( $me->Socket() ) or $me->Message("close failed: $!") and return 0;
     $me->State('DEAD');
+    $line =~ /^\+OK/i || return 0;
   }
   1;
 } # end Close
 
 
 #******************************************************************************
-#* 
+#*
 #******************************************************************************
 sub DESTROY
 {
@@ -305,7 +335,7 @@ sub DESTROY
 sub Connect
 {
   my ($me, $host, $port) = @_;
-  
+
   $host and $me->Host($host);
   $port and $me->Port($port);
 
@@ -315,37 +345,72 @@ sub Connect
 				 PeerPort  => $me->Port(),
 				 Proto     => "tcp",
 				 Type      => SOCK_STREAM,
+				 LocalAddr => $me->LocalAddr(),
 				 Timeout   => $me->{TIMEOUT} )
-    or 
+    or
       $me->Message( "could not connect socket [$me->{HOST}, $me->{PORT}]: $!" )
 	and
 	  return 0;
   $me->{SOCKET} = $s;
 
   $s->autoflush( 1 );
-  
+
   defined(my $msg = $me->_sockread()) or $me->Message("Could not read") and return 0;
   chomp $msg;
   $me->{BANNER}= $msg;
+
+  # add check for servers that return -ERR on connect (not in RFC1939)
+  $me->Message($msg);
+  $msg =~ /^\+OK/i || return 0;
+
   $me->{MESG_ID}= $1 if ($msg =~ /(<[\w\d\-\.]+\@[\w\d\-\.]+>)/);
   $me->Message($msg);
+
   $me->State('AUTHORIZATION');
-  
-  $me->User() and $me->Pass() and $me->Login();
-  
+  defined($me->User()) and defined($me->Pass()) and $me->Login();
+
 } # end Connect
 
 
 #******************************************************************************
-#* login to the POP server.  If the AUTH_MODE is set to APOP an the
-#* server supports it, it will use APOP.  Otherwise uid and password are
-#* sent in clear text.
+#* login to the POP server. If the AUTH_MODE is set to BEST, and the server
+#* appears to support APOP, it will try APOP, if that fails, then it will try
+#* SASL CRAM-MD5 if the server appears to support it, and finally PASS.
+#* If the AUTH_MODE is set to APOP, and the server appears to support APOP, it
+#* will use APOP or it will fail to log in. Likewise, for AUTH_MODE CRAM-MD5,
+#* no PASS-fallback is made. Otherwise password is sent in clear text.
 #******************************************************************************
 sub Login
 {
   my $me= shift;
-  if ($me->{AUTH_MODE} eq 'APOP' && $me->{MESG_ID}) { $me->Login_APOP(); }
-  else { $me->Login_Pass(); }
+
+  if ($me->{AUTH_MODE} eq 'BEST') {
+    my $retval;
+    if ($me->{MESG_ID}) {
+      $retval = $me->Login_APOP();
+      return($retval) if ($me->State eq 'TRANSACTION');
+    }
+    my $has_cram_md5 = 0;
+    foreach my $capa ($me->Capa()) {
+      $capa =~ /^SASL.*?\sCRAM-MD5\b/ and $has_cram_md5 = 1 and last;
+    }
+    if ($has_cram_md5) {
+      $retval = $me->Login_CRAM_MD5();
+      return($retval) if ($me->State() eq 'TRANSACTION');
+    }
+  }
+  elsif ($me->{AUTH_MODE} eq 'APOP') {
+    return(0) if (!$me->{MESG_ID});   # fail if the server does not support APOP
+    return($me->Login_APOP());
+  }
+  elsif ($me->{AUTH_MODE} eq 'CRAM-MD5') {
+    return($me->Login_CRAM_MD5());
+  }
+  elsif ($me->{AUTH_MODE} ne 'PASS') {
+    $me->Message("Programing error. AUTH_MODE (".$me->{AUTH_MODE}.") not BEST | APOP | CRAM-MD5 | PASS.");
+    return(0);
+  }
+  return($me->Login_Pass());
 }
 
 
@@ -354,20 +419,64 @@ sub Login
 #******************************************************************************
 sub Login_APOP
 {
-  require MD5;
-  
   my $me = shift;
-  my $s = $me->Socket();
-  my $hash= MD5->hexhash ($me->{MESG_ID} . $me->Pass);
-  
+
+  eval {
+    require Digest::MD5;
+  };
+  $@ and $me->Message("APOP failed: $@") and return 0;
+
+  my $hash = Digest::MD5::md5_hex($me->{MESG_ID} . $me->Pass());
+
+  $me->_checkstate('AUTHORIZATION', 'APOP') or return 0;
   $me->_sockprint( "APOP " , $me->User , ' ', $hash, $me->EOL );
   my $line = $me->_sockread();
   chomp $line;
   $me->Message($line);
-  $line =~ /^\+OK/ or $me->Message("APOP failed: $line") and $me->State('AUTHORIZATION')
-    and return 0;
+  $line =~ /^\+OK/ or $me->Message("APOP failed: $line") and return 0;
   $me->State('TRANSACTION');
-  
+
+  $me->POPStat() or return 0;
+}
+
+
+#******************************************************************************
+#* login to the POP server using CRAM-MD5 (RFC 2195) authentication.
+#******************************************************************************
+sub Login_CRAM_MD5
+{
+  my $me = shift;
+
+  eval {
+    require Digest::HMAC_MD5;
+    require MIME::Base64;
+  };
+  $@ and $me->Message("AUTH CRAM-MD5 failed: $@") and return 0;
+
+  $me->_checkstate('AUTHORIZATION', 'AUTH') or return 0;
+  $me->_sockprint('AUTH CRAM-MD5', $me->EOL());
+  my $line = $me->_sockread();
+  chomp $line;
+  $me->Message($line);
+
+  if ($line =~ /^\+ (.+)$/) {
+
+    my $hmac =
+      Digest::HMAC_MD5::hmac_md5_hex(MIME::Base64::decode($1), $me->Pass());
+    $me->_sockprint(MIME::Base64::encode($me->User() . " $hmac", $me->EOL()));
+
+    $line = $me->_sockread();
+    chomp $line;
+    $me->Message($line);
+    $line =~ /^\+OK/ or
+      $me->Message("AUTH CRAM-MD5 failed: $line") and return 0;
+
+  } else {
+    $me->Message("AUTH CRAM-MD5 failed: $line") and return 0;
+  }
+
+  $me->State('TRANSACTION');
+
   $me->POPStat() or return 0;
 }
 
@@ -378,25 +487,24 @@ sub Login_APOP
 sub Login_Pass
 {
   my $me = shift;
-  my $s = $me->Socket();
+
+  $me->_checkstate('AUTHORIZATION', 'USER') or return 0;
   $me->_sockprint( "USER " , $me->User() , $me->EOL );
   my $line = $me->_sockread();
   chomp $line;
   $me->Message($line);
-  $line =~ /^\+/ or $me->Message("USER failed: $line") and $me->State('AUTHORIZATION')
-    and return 0;
-  
+  $line =~ /^\+/ or $me->Message("USER failed: $line") and return 0;
+
   $me->_sockprint( "PASS " , $me->Pass() , $me->EOL );
   $line = $me->_sockread();
   chomp $line;
   $me->Message($line);
-  $line =~ /^\+OK/ or $me->Message("PASS failed: $line") and $me->State('AUTHORIZATION')
-    and return 0;
-  
+  $line =~ /^\+OK/ or $me->Message("PASS failed: $line") and return 0;
+
   $me->State('TRANSACTION');
 
   $me->POPStat() or return 0;
-  
+
 } # end Login
 
 
@@ -412,21 +520,23 @@ sub Head
   my $lines = shift;
   $lines ||= 0;
   $lines =~ /\d+/ || ($lines = 0);
+
   my $header = '';
-  my $s = $me->Socket();
-  
+
+  $me->_checkstate('TRANSACTION', 'TOP') or return;
   $me->_sockprint( "TOP $num $lines", $me->EOL );
   my $line = $me->_sockread();
   chomp $line;
   $line =~ /^\+OK/ or $me->Message("Bad return from TOP: $line") and return;
   $line =~ /^\+OK (\d+) / and my $buflen = $1;
-  
-  do {
+
+  while (1) {
     $line = $me->_sockread();
-    #    $line =~ /^\s*$|^\.\s*$/ or $header .= $line;
-    $line =~ /^\.\s*$/ or $header .= $line;
-  } until $line =~ /^\.\s*$/;
-  
+    last if $line =~ /^\.\s*$/;
+    $line =~ s/^\.\././;
+    $header .= $line;
+  }
+
   return wantarray ? split(/\r?\n/, $header) : $header;
 } # end Head
 
@@ -439,24 +549,57 @@ sub HeadAndBody
   my $me = shift;
   my $num = shift;
   my $mandb = '';
-  my $s = $me->Socket();
-  
+
+  $me->_checkstate('TRANSACTION', 'RETR') or return;
   $me->_sockprint( "RETR $num", $me->EOL );
   my $line = $me->_sockread();
   chomp $line;
   $line =~ /^\+OK/ or $me->Message("Bad return from RETR: $line") and return;
   $line =~ /^\+OK (\d+) / and my $buflen = $1;
-  
-  do {
+
+  while (1) {
     $line = $me->_sockread();
-    $line =~ /^\.\s*$/ or $mandb .= $line;
-  } until $line =~ /^\.\s*$/;
-  
+    last if $line =~ /^\.\s*$/;
+    # convert any '..' at the start of a line to '.'
+    $line =~ s/^\.\././;
+    $mandb .= $line;
+  }
+
   return wantarray ? split(/\r?\n/, $mandb) : $mandb;
-  
+
 } # end HeadAndBody
 
 
+#******************************************************************************
+#* get the head and body of a message and write it to a file handle.
+#* Sends the raw data: does no CR/NL stripping or mapping.
+#******************************************************************************
+sub HeadAndBodyToFile
+{
+  my $me = shift;
+  my $fh = shift;
+  my $num = shift;
+  my $body = '';
+
+  $me->_checkstate('TRANSACTION', 'RETR') or return;
+  $me->_sockprint( "RETR $num", $me->EOL );
+  my $line = $me->_sockread();
+  chomp $line;
+  $line =~ /^\+OK/ or $me->Message("Bad return from RETR: $line") and return 0;
+  $line =~ /^\+OK (\d+) / and my $buflen = $1;
+
+  while (1) {
+    $line = $me->_sockread();
+    last if $line =~ /^\.\s*$/;
+    # convert any '..' at the start of a line to '.'
+    $line =~ s/^\.\././;
+    print $fh $line;
+  }
+  return 1;
+} # end BodyToFile
+
+
+
 #******************************************************************************
 #* get the body of a message
 #******************************************************************************
@@ -465,41 +608,80 @@ sub Body
   my $me = shift;
   my $num = shift;
   my $body = '';
-  my $s = $me->Socket();
-  
+
+  $me->_checkstate('TRANSACTION', 'RETR') or return;
   $me->_sockprint( "RETR $num", $me->EOL );
   my $line = $me->_sockread();
   chomp $line;
   $line =~ /^\+OK/ or $me->Message("Bad return from RETR: $line") and return;
   $line =~ /^\+OK (\d+) / and my $buflen = $1;
-  
+
   # skip the header
   do {
     $line = $me->_sockread();
   } until $line =~ /^\s*$/;
-  
+
+  while (1) {
+    $line = $me->_sockread();
+    last if $line =~ /^\.\s*$/;
+    # convert any '..' at the start of a line to '.'
+    $line =~ s/^\.\././;
+    $body .= $line;
+  }
+
+  return wantarray ? split(/\r?\n/, $body) : $body;
+
+} # end Body
+
+
+#******************************************************************************
+#* get the body of a message and write it to a file handle.  Sends the raw data:
+#* does no CR/NL stripping or mapping.
+#******************************************************************************
+sub BodyToFile
+{
+  my $me = shift;
+  my $fh = shift;
+  my $num = shift;
+  my $body = '';
+
+  $me->_checkstate('TRANSACTION', 'RETR') or return;
+  $me->_sockprint( "RETR $num", $me->EOL );
+  my $line = $me->_sockread();
+  chomp $line;
+  $line =~ /^\+OK/ or $me->Message("Bad return from RETR: $line") and return;
+  $line =~ /^\+OK (\d+) / and my $buflen = $1;
+
+  # skip the header
   do {
     $line = $me->_sockread();
-    $line =~ /^\.\s*$/ or $body .= $line;
-  } until $line =~ /^\.\s*$/;
-  
-  return wantarray ? split(/\r?\n/, $body) : $body;
-  
-} # end Body
+  } until $line =~ /^\s*$/;
+
+  while (1) {
+    $line = $me->_sockread();
+    chomp $line;
+    last if $line =~ /^\.\s*$/;
+    # convert any '..' at the start of a line to '.'
+    $line =~ s/^\.\././;
+    print $fh $line, "\n";
+  }
+} # end BodyToFile
+
 
 
 #******************************************************************************
 #* handle a STAT command - returns the number of messages in the box
 #******************************************************************************
-sub POPStat {
+sub POPStat
+{
   my $me = shift;
-  my $s = $me->Socket();
-  
+
+  $me->_checkstate('TRANSACTION', 'STAT') or return -1;
   $me->_sockprint( "STAT", $me->EOL );
   my $line = $me->_sockread();
   $line =~ /^\+OK/ or $me->Message("STAT failed: $line") and return -1;
   $line =~ /^\+OK (\d+) (\d+)/ and $me->Count($1), $me->Size($2);
-  
+
   return $me->Count();
 }
 
@@ -512,14 +694,15 @@ sub List {
   my $num = shift || '';
   my $CMD = shift || 'LIST';
   $CMD=~ tr/a-z/A-Z/;
-  
-  my $s = $me->Socket();
+
   $me->Alive() or return;
-  
+
   my @retarray = ();
   my $ret = '';
-  
-  $me->_sockprint( "$CMD $num", $me->EOL );
+
+  $me->_checkstate('TRANSACTION', $CMD) or return;
+  $me->_sockprint($CMD, $num ? " $num" : '', $me->EOL());
+
   my $line = $me->_sockread();
   $line =~ /^\+OK/ or $me->Message("$line") and return;
   if ($num) {
@@ -537,6 +720,40 @@ sub List {
   }
 }
 
+#******************************************************************************
+#* issue the LIST command, but return results in an indexed array.
+#******************************************************************************
+sub ListArray {
+  my $me = shift;
+  my $num = shift || '';
+  my $CMD = shift || 'LIST';
+  $CMD=~ tr/a-z/A-Z/;
+
+  $me->Alive() or return;
+
+  my @retarray = ();
+  my $ret = '';
+
+  $me->_checkstate('TRANSACTION', $CMD) or return;
+  $me->_sockprint($CMD, $num ? " $num" : '', $me->EOL());
+  my $line = $me->_sockread();
+  $line =~ /^\+OK/ or $me->Message("$line") and return;
+  if ($num) {
+    $line =~ s/^\+OK\s*//;
+    return $line;
+  }
+  while( defined( $line = $me->_sockread() ) ) {
+    $line =~ /^\.\s*$/ and last;
+    $ret .= $line;
+    chomp $line;
+    my ($num, $uidl) = split ' ', $line;
+    $retarray[$num] = $uidl;
+  }
+  if ($ret) {
+    return wantarray ? @retarray : $ret;
+  }
+}
+
 
 #******************************************************************************
 #* retrieve the given message number - uses HeadAndBody
@@ -545,18 +762,26 @@ sub Retrieve {
   return HeadAndBody( @_ );
 }
 
+#******************************************************************************
+#* retrieve the given message number to the given file handle- uses
+#* HeadAndBodyToFile
+#******************************************************************************
+sub RetrieveToFile {
+  return HeadAndBodyToFile( @_ );
+}
+
 
 #******************************************************************************
 #* implement the LAST command - see the rfc (1081) OBSOLETED by RFC
 #******************************************************************************
-sub Last {
+sub Last
+{
   my $me = shift;
-  
-  my $s = $me->Socket();
-  
+
+  $me->_checkstate('TRANSACTION', 'LAST') or return;
   $me->_sockprint( "LAST", $me->EOL );
   my $line = $me->_sockread();
-  
+
   $line =~ /\+OK (\d+)\D*$/ and return $1;
 }
 
@@ -564,13 +789,14 @@ sub Last {
 #******************************************************************************
 #* reset the deletion stat
 #******************************************************************************
-sub Reset {
+sub Reset
+{
   my $me = shift;
-  
-  my $s = $me->Socket();
+
+  $me->_checkstate('TRANSACTION', 'RSET') or return;
   $me->_sockprint( "RSET", $me->EOL );
   my $line = $me->_sockread();
-  $line =~ /\+OK .*$/ and return 1;
+  $line =~ /^\+OK/ and return 1;
   return 0;
 }
 
@@ -581,12 +807,12 @@ sub Reset {
 sub Delete {
   my $me = shift;
   my $num = shift || return;
-  
-  my $s = $me->Socket();
+
+  $me->_checkstate('TRANSACTION', 'DELE') or return;
   $me->_sockprint( "DELE $num",  $me->EOL );
   my $line = $me->_sockread();
   $me->Message($line);
-  $line =~ /^\+OK / && return 1;
+  $line =~ /^\+OK/ && return 1;
   return 0;
 }
 
@@ -594,17 +820,18 @@ sub Delete {
 #******************************************************************************
 #* UIDL - submitted by Dion Almaer (dion@member.com)
 #******************************************************************************
-sub Uidl {
+sub Uidl
+{
   my $me = shift;
   my $num = shift || '';
-  
-  my $s = $me->Socket();
+
   $me->Alive() or return;
-  
+
   my @retarray = ();
   my $ret = '';
-  
-  $me->_sockprint( "UIDL $num", $me->EOL );
+
+  $me->_checkstate('TRANSACTION', 'UIDL') or return;
+  $me->_sockprint('UIDL', $num ? " $num" : '', $me->EOL());
   my $line = $me->_sockread();
   $line =~ /^\+OK/ or $me->Message($line) and return;
   if ($num) {
@@ -624,6 +851,54 @@ sub Uidl {
 }
 
 
+#******************************************************************************
+#* CAPA - query server capabilities, see RFC 2449
+#******************************************************************************
+sub Capa {
+
+  my $me = shift;
+
+  # no state check here, all are allowed
+
+  $me->Alive() or return;
+
+  my @retarray = ();
+  my $ret = '';
+
+  $me->_sockprint('CAPA', $me->EOL());
+
+  my $line = $me->_sockread();
+  $line =~ /^\+OK/ or $me->Message($line) and return;
+
+  while(defined($line = $me->_sockread())) {
+    $line =~ /^\.\s*$/ and last;
+    $ret .= $line;
+    chomp $line;
+    push(@retarray, $line);
+  }
+
+  if ($ret) {
+    return wantarray ? @retarray : $ret;
+  }
+}
+
+
+#*****************************************************************************
+#* Check the state before issuing a command
+#*****************************************************************************
+sub _checkstate
+{
+  my ($me, $state, $cmd) = @_;
+  my $currstate = $me->State();
+  if ($currstate ne $state) {
+    $me->Message("POP3 command $cmd may be given only in the '$state' state " .
+                 "(current state is '$currstate').");
+    return 0;
+  } else {
+    return 1;
+  }
+}
+
 
 #*****************************************************************************
 #* funnel all read/write through here to allow easier debugging
@@ -633,7 +908,10 @@ sub _sockprint
 {
   my $me = shift;
   my $s = $me->Socket();
-  $me->Debug and carp "POP3 -> ", @_;
+  $me->Debug and Carp::carp "POP3 -> ", @_;
+  my $outline = "@_";
+  chomp $outline;
+  push(@{$me->{tranlog}}, $outline);
   print $s @_;
 }
 
@@ -641,7 +919,17 @@ sub _sockread
 {
   my $me = shift;
   my $line = $me->Socket()->getline();
-  $me->Debug and carp "POP3 <- ", $line;
+
+  # Macs seem to leave CR's or LF's sitting on the socket.  This
+  # removes them.
+  $me->{STRIPCR} && ($line =~ s/^[\r]+//);
+
+  $me->Debug and Carp::carp "POP3 <- ", $line;
+  $line =~ /^[\\+\\-](OK|ERR)/i && do {
+    my $l = $line;
+    chomp $l;
+    push(@{$me->{tranlog}}, $l);
+  };
   return $line;
 }
 
@@ -679,7 +967,7 @@ Mail::POP3Client - Perl 5 module to talk to a POP3 (RFC1939) server
   $pop2 = new Mail::POP3Client( HOST  => "pop3.otherdo.main" );
   $pop2->User( "somebody" );
   $pop2->Pass( "doublesecret" );
-  $pop2->Connect() || die $pop2->Message();
+  $pop2->Connect() >= 0 || die $pop2->Message();
   $pop2->Close();
 
 =head1 DESCRIPTION
@@ -693,9 +981,9 @@ Here is a simple example to list out the From: and Subject: headers in
 your remote mailbox:
 
   #!/usr/local/bin/perl
-  
+
   use Mail::POP3Client;
-  
+
   $pop = new Mail::POP3Client( USER     => "me",
 			       PASSWORD => "mypassword",
 			       HOST     => "pop3.do.main" );
@@ -716,9 +1004,10 @@ New style (shown with defaults):
                          PASSWORD  => "",
                          HOST      => "pop3",
                          PORT      => 110,
-                         AUTH_MODE => 'PASS',
+                         AUTH_MODE => 'BEST',
                          DEBUG     => 0,
                          TIMEOUT   => 60,
+                         LOCALADDR => 'xxx.xxx.xxx.xxx[:xx]',
                        );
 
 =over 4
@@ -739,11 +1028,14 @@ PORT is the POP server port (default = 110)
 DEBUG - any non-null, non-zero value turns on debugging (default = 0)
 
 =item *
-AUTH_MODE - pass 'APOP' to attempt APOP (MD5) authorization. (default is 'PASS')
+AUTH_MODE - pass 'APOP' to force APOP (MD5) authorization. (default is 'BEST')
 
 =item *
 TIMEOUT - set a timeout value for socket operations (default = 60)
 
+=item *
+LOCALADDR - allow selecting a local inet address to use
+
 =back
 
 =head1 METHODS
@@ -756,7 +1048,7 @@ commands return multiple lines as an array in an array context.
 =over 8
 
 =item I<new>( USER => 'user', PASSWORD => 'password', HOST => 'host',
-              PORT => 110, DEBUG => 0, AUTH_MODE => 'PASS', TIMEOUT => 60 )
+              PORT => 110, DEBUG => 0, AUTH_MODE => 'BEST', TIMEOUT => 60 )
 
 Construct a new POP3 connection with this.  You should use the
 hash-style constructor.  B<The old positional constructor is
@@ -771,19 +1063,28 @@ new will attempt to Connect to and Login to the POP3 server if you
 supply a USER and PASSWORD.  If you do not supply them in the
 constructor, you will need to call Connect yourself.
 
-The valid values for AUTH_MODE are 'PASS' and 'APOP'.  APOP implies
-that an MD5 checksum will be used instrad of passing your password in
-cleartext.  However, B<if the server does not support APOP, the
-cleartext method will be used.  Be careful.>
+The valid values for AUTH_MODE are 'BEST', 'PASS', 'APOP' and 'CRAM-MD5'.
+BEST says to try APOP if the server appears to support it and it can be
+used to successfully log on, next try similarly with CRAM-MD5, and finally
+revert to PASS. APOP and CRAM-MD5 imply that an MD5 checksum will be
+used instead of sending your password in cleartext.  However,
+B<if the server does not claim to support APOP or CRAM-MD5,
+the cleartext method will be used. Be careful.> There are a few
+servers that will send a timestamp in the banner greeting, but APOP
+will not work with them (for instance if the server does not know your
+password in cleartext).  If you think your authentication information
+is correct, run in DEBUG mode and look for errors regarding
+authorization.  If so, then you may have to use 'PASS' for that server.
+The same applies to CRAM-MD5, too.
 
-If you enable debugging with DEBUG => 1, messages about command will
-go to STDERR.
+If you enable debugging with DEBUG => 1, socket traffic will be echoed
+to STDERR.
 
 Another warning, it's impossible to differentiate between a timeout
 and a failure.
 
 
-=item I<Head>( MESSAGE_NUMBER )
+=item I<Head>( MESSAGE_NUMBER [, PREVIEW_LINES ] )
 
 Get the headers of the specified message, either as an array or as a
 string, depending on context.
@@ -798,7 +1099,17 @@ Dennis Moroney <dennis@hub.iwl.net>.
 Get the body of the specified message, either as an array of lines or
 as a string, depending on context.
 
-=item I<HeadAndBody>( MESSAGE_NUMBER [, PREVIEW_LINES ] )
+=item I<BodyToFile>( FILE_HANDLE, MESSAGE_NUMBER )
+
+Get the body of the specified message and write it to the given file handle.
+my $fh = new IO::Handle();
+$fh->fdopen( fileno( STDOUT ), "w" );
+$pop->BodyToFile( $fh, 1 );
+
+Does no stripping of NL or CR.
+
+
+=item I<HeadAndBody>( MESSAGE_NUMBER )
 
 Get the head and body of the specified message, either as an array of
 lines or as a string, depending on context.
@@ -807,22 +1118,35 @@ lines or as a string, depending on context.
 
 =item Example
 
-foreach ( $pop->HeadAndBody( 1, 10 ) )
+foreach ( $pop->HeadAndBody( 1 ) )
    print $_, "\n";
 
-prints out a preview of each message, with the full header and the
-first 10 lines of the message (if supported by the POP3 server).
+prints out the complete text of message 1.
 
 =back
 
+=item I<HeadAndBodyToFile>( FILE_HANDLE, MESSAGE_NUMBER )
+
+Get the head and body of the specified message and write it to the given file handle.
+my $fh = new IO::Handle();
+$fh->fdopen( fileno( STDOUT ), "w" );
+$pop->HeadAndBodyToFile( $fh, 1 );
+
+Does no stripping of NL or CR.
+
+
 =item I<Retrieve>( MESSAGE_NUMBER )
 
 Same as HeadAndBody.
 
+=item I<RetrieveToFile>( FILE_HANDLE, MESSAGE_NUMBER )
+
+Same as HeadAndBodyToFile.
+
 =item I<Delete>( MESSAGE_NUMBER )
 
 Mark the specified message number as DELETED.  Becomes effective upon
-QUIT.  Can be reset with a Reset message.
+QUIT (invoking the Close method).  Can be reset with a Reset message.
 
 =item I<Connect>
 
@@ -852,7 +1176,8 @@ Set/Return the number of remote messages.  Set during Login.
 
 =item I<Message>
 
-The last status message received from the server.
+The last status message received from the server or a message
+describing any problem encountered.
 
 =item I<State>
 
@@ -863,13 +1188,39 @@ The internal state of the connection: DEAD, AUTHORIZATION, TRANSACTION.
 Return the results of a POP3 STAT command.  Sets the size of the
 mailbox.
 
-=item I<List>
+=item I<List>([message_number])
 
-Return a list of sizes of each message.
+Returns the size of the given message number when called with an
+argument using the following format:
+
+   <message_number> <size_in_bytes>
+
+If message_number is omitted, List behaves the same as ListArray,
+returning an indexed array of the sizes of each message in the same
+format.
+
+You can parse the size in bytes using split:
+ ($msgnum, $size) = split ' ', $pop -> List( n )
+
+
+=item I<ListArray>
+
+Return a list of sizes of each message.  This returns an indexed
+array, with each message number as an index (starting from 1) and the
+value as the next entry on the line.  Beware that some servers send
+additional info for each message for the list command.  That info may
+be lost.
 
 =item I<Uidl>( [MESSAGE_NUMBER] )
 
-Return the unique ID for the given message (or all of them).
+Return the unique ID for the given message (or all of them).  Returns
+an indexed array with an entry for each valid message number.
+Indexing begins at 1 to coincide with the server's indexing.
+
+=item I<Capa>
+
+Query server capabilities, as described in RFC 2449. Returns the
+capabilities in an array. Valid in all states.
 
 =item I<Last>
 
@@ -901,9 +1252,15 @@ Set/Return the current port number.
 
 =back
 
+=head1 REQUIREMENTS
+
+This module does not have mandatory requirements for modules that are not part
+of the standard Perl distribution. However, APOP needs need Digest::MD5 and
+CRAM-MD5 needs Digest::HMAC_MD5 and MIME::Base64.
+
 =head1 AUTHOR
 
-Sean Dowd <dowd@home.com>
+Sean Dowd <pop3client@dowds.net>
 
 =head1 CREDITS
 
@@ -912,6 +1269,14 @@ Based loosely on News::NNTPClient by Rodger Anderson
 
 =head1 SEE ALSO
 
-perl(1).
+perl(1)
+
+the Digest::MD5 manpage, the Digest::HMAC_MD5 manpage, the MIME::Base64 manpage
+
+RFC 1939: Post Office Protocol - Version 3
+
+RFC 2195: IMAP/POP AUTHorize Extension for Simple Challenge/Response
+
+RFC 2449: POP3 Extension Mechanism
 
 =cut
